import argparse
import os
import sys
from MainShortcuts2 import ms


def import_example():
  argp = argparse.ArgumentParser("ms2-import_example", description="код для импорта MainShortcuts2")
  argp.parse_args()
  print("from MainShortcuts2 import ms")


def nano_json(args: argparse.Namespace = None):
  ms.utils.check_programs("nano")
  import subprocess
  if args is None:
    argp = argparse.ArgumentParser("nano-json", description="форматирование JSON файлов и редактирование в GNU NANO")
    argp.add_argument("files", nargs="+", help="пути к файлам JSON")
    argp.add_argument("--nano-help", action="store_true", help="показать помощь nano")
    argp.add_argument("-e", "--encoding", default="utf-8", help="кодировка файлов")
    argp.add_argument("-f", "--rcfile", help="использовать только этот файл для настройки nano")
    argp.add_argument("-i", "--indent", default=2, type=int, help="кол-во пробелов для отступа")
    argp.add_argument("-m", "--mode", choices=ms.json.MODES_ALL, default="p", help="режим сохранения редактирования")
    argp.add_argument("-s", "--sort", action="store_true", help="сортировать ключи словаря")
    argp.add_argument("-u", "--no-escape", action="store_false", help="не использовать Unicode Escape")
    args = argp.parse_args()
  if args.nano_help:
    return subprocess.call(["nano", "--help"])
  write_kw = {}
  write_kw["encoding"] = args.encoding
  write_kw["ensure_ascii"] = False
  write_kw["indent"] = args.indent
  write_kw["mode"] = "print"
  write_kw["sort_keys"] = args.sort
  nano_args = ["nano"]
  if args.rcfile:
    nano_args += ["--rcfile", args.rcfile]
  nano_args += args.files
  for i in args.files:
    try:
      data = ms.json.read(i, encoding=args.encoding)
      ms.json.write(i, data, **write_kw)
    except Exception as err:
      print(err, file=sys.stderr)
  subprocess.call(nano_args)
  write_kw["ensure_ascii"] = args.no_escape
  write_kw["indent"] = None if args.mode in ms.json.MODES_COMPRESS else args.indent
  write_kw["mode"] = args.mode
  for i in args.files:
    try:
      data = ms.json.read(i, encoding=args.encoding)
      ms.json.write(i, data, **write_kw)
    except Exception as err:
      print(err, file=sys.stderr)


def _check_nginx() -> int:
  import subprocess
  with subprocess.Popen(["nginx", "-t"], stderr=subprocess.PIPE) as p:
    code = p.wait()
    if code != 0:
      sys.stderr.buffer.write(p.stderr.read())
  return code


def nginx_reload():
  ms.utils.check_programs("nginx")
  import subprocess
  argp = argparse.ArgumentParser("nginx-reload", description="проверка конфига и перезагрузка Nginx")
  argp.parse_args()
  code = _check_nginx()
  if code != 0:
    sys.exit(code)
  sys.exit(subprocess.call(["nginx", "-s", "reload"]))


def nginx_restart():
  ms.utils.check_programs("nginx", "systemctl")
  import subprocess
  argp = argparse.ArgumentParser("nginx-reload", description="проверка конфига и перезапуск Nginx через systemctl")
  argp.parse_args()
  code = _check_nginx()
  if code != 0:
    sys.exit(code)
  sys.exit(subprocess.call(["systemctl", "restart", "nginx"]))


def ln(args=None):
  if args is None:
    argp = argparse.ArgumentParser("ms2-ln", description="создание символьной ссылки с абсолютным путём к файлу")
    argp.add_argument("-f", "--force", action="store_true", help="принудительное создание ссылки")
    argp.add_argument("file", help="путь к файлу, на который будет указывать ссылка")
    argp.add_argument("link", help="путь к ссылке для создания")
    args = argp.parse_args()
  if args.force:
    ms.path.delete(args.link)
  file = os.path.realpath(args.file)
  os.symlink(file, args.link, os.path.isdir(file))



import os
import sys
from . import _module_info
from datetime import datetime
from logging import Logger
from time import time
from typing import Union
try:
  from datetime import timezone
except Exception:
  timezone = None


def _get_main_file():
  def func():
    if getattr(sys, "frozen", False):
      return sys.executable
    if "__main__" in sys.modules:
      if hasattr(sys.modules["__main__"], "__file__"):
        return sys.modules["__main__"].__file__
  result = func()
  if result is None:
    return
  return os.path.abspath(result).replace("\\", "/")


class NoLogger(Logger):
  def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1):
    pass


class MS2:
  version = _module_info.version

  def __init__(self, *,
               __file__: str = None,
               __name__: str = None,
               logger: Logger = None,
               **kw):
    self._advanced = None
    self._any2json = None
    self._cfg = None
    self._dict = None
    self._dir = None
    self._file = None
    self._json = None
    self._list = None
    self._ms2app = None
    self._ms2hash = None
    self._path = None
    self._proc = None
    self._regex = None
    self._special_chars = None
    self._str = None
    self._term = None
    self._types = None
    self._utils = None
    self._win = None
    self.args: list[str] = sys.argv
    self.encoding: str = "utf-8"
    self.env: dict[str, str] = os.environ
    self.import_code: str = "from MainShortcuts2 import ms\nms.prog_file,ms.prog_name=__file__,__name__\nms.reload()"
    self.log: Logger = NoLogger("MainShortcuts2") if logger is None else logger
    self.MAIN_FILE: Union[None, str] = _get_main_file()
    self.MAIN_DIR: Union[None, str] = None if self.MAIN_FILE is None else os.path.dirname(self.MAIN_FILE)
    self.prog_dir: Union[None, str] = None
    self.prog_file: Union[None, str] = __file__
    self.prog_name: Union[None, str] = __name__
    self.use_tmp_file: bool = False
    self.reload()

  def reload(self):
    self.pid = os.getpid()
    if not self.prog_file is None:
      self.prog_dir = os.path.dirname(self.prog_file)
    if not self.prog_name is None:
      self.imported: bool = self.prog_name != "__main__"
      self.running: bool = self.prog_name == "__main__"

  @property
  def credits(self) -> str:
    line = "------------------------------------------------"
    l = []
    l.append(line)
    l.append(("MainShortcuts2 %s" % self.version).center(len(line)))
    l.append("https://github.com/MainPlay-TG/MainShortcuts2.py")
    l.append(line)
    l.append("Разработчик: MainPlay TG")
    l.append("- Telegram - https://t.me/MainPlay_TG")
    l.append("- GitHub - https://github.com/MainPlay-TG")
    l.append("- Пожертвования - https://www.donationalerts.com/r/mainplay_tg (не знаю зачем, всё равно никто и копейки не даст)")
    l.append(line)
    l.append("Тестеры:")
    l.append("- MainPlay TG (https://t.me/MainPlay_TG)")
    l.append("- Кларк (https://t.me/Klark_Anthrofox)")
    l.append("- FOX (https://t.me/black_fox404)")
    l.append("- Ты (напиши мне в Telegram и в ближайшем обновлении тут будет твоя ссылка)")
    l.append(line)
    l.append("Спасибо за использование моей библиотеки".center(len(line)))
    return "\n".join(l)

  @property
  def advanced(self):
    if self._advanced is None:
      from . import advanced
      self._advanced = advanced
    return self._advanced

  @property
  def any2json(self):
    if self._any2json is None:
      from . import any2json
      self._any2json = any2json
    return self._any2json

  @property
  def cfg(self):
    if self._cfg is None:
      from .cfg import cfg
      self._cfg = cfg
    return self._cfg

  @property
  def dict(self):
    if self._dict is None:
      from . import dict
      self._dict = dict
    return self._dict

  @property
  def dir(self):
    if self._dir is None:
      from . import dir
      self._dir = dir
    return self._dir

  @property
  def file(self):
    if self._file is None:
      from . import file
      self._file = file
    return self._file

  @property
  def json(self):
    if self._json is None:
      from . import json
      self._json = json
    return self._json

  @property
  def list(self):
    if self._list is None:
      from . import list
      self._list = list
    return self._list

  @property
  def ms2app(self):
    if self._ms2app is None:
      from . import ms2app
      self._ms2app = ms2app
    return self._ms2app

  @property
  def ms2hash(self):
    if self._ms2hash is None:
      from . import ms2hash
      self._ms2hash = ms2hash
    return self._ms2hash

  @property
  def path(self):
    if self._path is None:
      from . import path
      self._path = path
    return self._path

  @property
  def proc(self):
    if self._proc is None:
      from . import proc
      self._proc = proc
    return self._proc

  @property
  def regex(self):
    if self._regex is None:
      from . import regex
      self._regex = regex
    return self._regex

  @property
  def special_chars(self):
    if self._special_chars is None:
      from . import special_chars
      self._special_chars = special_chars
    return self._special_chars

  @property
  def str(self):
    if self._str is None:
      from . import str
      self._str = str
    return self._str

  @property
  def term(self):
    if self._term is None:
      from . import term
      self._term = term
    return self._term

  @property
  def types(self):
    if self._types is None:
      from . import types
      self._types = types
    return self._types

  @property
  def utils(self):
    if self._utils is None:
      from . import utils
      self._utils = utils
    return self._utils

  @property
  def win(self):
    if self._win is None:
      from . import win
      self._win = win
    return self._win

  class ObjectBase(object):
    def __new__(cls, *a, **b):
      self = object.__new__(cls)
      return self

    def __init__(self, *a, **b):
      pass

    def __bool__(self):
      return True

    def __enter__(self):
      return self

    def __exit__(self, *a):
      if callable(getattr(self, "close", None)):
        self.close()

    def __repr__(self, args_string="..."):
      cls = type(self)
      return "%s.%s(%s)" % (cls.__module__, cls.__name__, args_string)

    @classmethod
    def _from_kw(cls, kw):
      return cls(**kw)

  @property
  def now(self) -> float:
    """Текущее локальное время (`timestamp`)"""
    return time()

  @property
  def now_dt(self) -> datetime:
    """Текущее локальное время (`datetime`)"""
    return datetime.fromtimestamp(time())

  @property
  def utcnow(self) -> float:
    """Текущее время по UTC (`timestamp`)"""
    return self.utcnow_dt.timestamp()
  if timezone is None:
    @property
    def utcnow_dt(self) -> datetime:
      """Текущее время по UTC (`datetime`)"""
      return datetime.utcfromtimestamp(time())
  else:
    @property
    def utcnow_dt(self) -> datetime:
      """Текущее время по UTC (`datetime`)"""
      return datetime.fromtimestamp(time(), timezone.utc)


ms = MS2()



"""Различные утилиты, требующие сторонних модулей"""
import builtins
import os
import sys
from .core import ms
from functools import wraps
from typing import *
cache = {}


def return_False(*a, **b):
  """Вернуть False при любых аргументах"""
  return False


def return_None(*a, **b):
  """Вернуть None при любых аргументах"""
  return None


def return_True(*a, **b):
  """Вернуть True при любых аргументах"""
  return True


class MiddlewareBase:
  """middleware для функции"""

  def __init__(self, func):
    self._init(func)

  def _init(self, func):
    self._args = None
    self._exception = None
    self._kwargs = None
    self._result = None
    self._time = None
    self._traceback = None
    self.completed_with_exc: Union[None, bool] = None
    self.completed: bool = False
    self.func = func
    self.launched: bool = False
    if not hasattr(self, "ignore_exceptions"):
      self.ignore_exceptions: bool = False

  def _check_completed(self):
    if not self.completed:
      raise RuntimeError("The function has not yet been completed")

  def _check_launched(self):
    if not self.launched:
      raise RuntimeError("The function has not yet been launched")

  def _run(self, args, kwargs):
    from time import time
    if self.launched:
      raise RuntimeError("You can only start a function once once")
    self._args = args
    self._kwargs = kwargs
    self.launched = True
    try:
      self.before()
    except Exception:
      if not self.ignore_exceptions:
        raise
    started_at = time()
    try:
      self._result = self.func(*self.args, **self.kwargs)
      self.completed_with_exc = False
    except Exception as e:
      from traceback import format_exc
      self._exception = e
      self._traceback = format_exc()
      self.completed_with_exc = True
    self._time = time() - started_at
    self.completed = True
    try:
      self.after()
    except Exception:
      if not self.ignore_exceptions:
        raise
    return self.result

  @property
  def args(self) -> tuple:
    """Аргументы, переданные функции"""
    self._check_launched()
    return self._args

  @args.setter
  def args(self, value: Iterable):
    if value is None:
      value = []
    self._args = tuple(value)

  @property
  def exception(self) -> None | Exception:
    """Исключение, возникшее в процессе работы функции"""
    self._check_completed()
    return self._exception

  @property
  def kwargs(self) -> dict[str, Any]:
    """Именованные аргументы, переданные функции"""
    self._check_launched()
    return self._kwargs

  @kwargs.setter
  def kwargs(self, value: dict[str, Any]):
    if value is None:
      value = {}
    for k in value.keys():
      assert type(k) == str
    self._kwargs = value

  @property
  def result(self) -> Any:
    """Результат работы функции"""
    self._check_completed()
    if self.completed_with_exc:
      raise self.exception  # type: ignore
    return self._result

  @property
  def time(self) -> float:
    """Продолжительность работы функции (сек)"""
    self._check_completed()
    return self._time

  @property
  def traceback(self) -> None | str:
    self._check_completed()
    return self._traceback

  def before(self):
    """Перед запуском функции. Можно обработать/изменить аргументы"""
    pass

  def after(self):
    """После запуска функции. Можно обработать результат или исключение"""
    pass


def args2kwargs(func: Callable, args: Iterable = (), kwargs: dict[str, Any] = {}) -> tuple[tuple, dict[str, Any]]:
  """Преобразовать `args` в `kwargs` | `inspect`"""
  import inspect
  kw = kwargs.copy()
  args = list(args)
  spec = inspect.getfullargspec(func)
  for i in inspect.signature(func).parameters:
    if i != spec.varargs:
      if i != spec.varkw:
        if not i in kw:
          kw[i] = args.pop(0)
  if len(args) > 0:
    if spec.varargs != None:
      raise TypeError("Too many arguments")
  return tuple(args), kw


async def async_download_file(url: str, path: str, *, cb_end=return_None, cb_progress=return_None, cb_start=return_None, chunk_size: int = 1024, delete_on_error: bool = True, **kw) -> int:
  """Асинхронная функция для скачивания файла | `aiohttp`"""
  kw.setdefault("method", "GET")
  kw["url"] = url
  async with async_request(**kw) as resp:  # type: ignore
    if callable(getattr(path, "write", None)):
      f: IO[bytes] = path
    else:
      f = open(path, "wb")
    with f:
      size = 0
      await cb_start(f, resp, size)
      try:
        async for chunk in resp.content.iter_chunked(chunk_size):
          size += f.write(chunk)
          await cb_progress(f, resp, size)
      except:
        if delete_on_error:
          if os.path.isfile(path):
            os.remove(path)
        raise
      cb_end(f, resp, size)
  return size


async def async_request(method: str, url: str, *, ignore_status: bool = False, session=None, **kw):
  """Асинхронный HTTP запрос | `aiohttp`"""
  from importlib import import_module
  aiohttp = import_module("aiohttp")
  # import aiohttp
  resp = None
  if isinstance(method, aiohttp.ClientResponse):
    resp = method
  if isinstance(url, aiohttp.ClientResponse):
    if not resp is None:
      raise TypeError("Only one argument can be `Response`")
    resp = url
  if resp is None:
    if session is None:
      session = aiohttp.ClientSession()
    kw["method"] = method
    kw["url"] = url
    resp = await session.request(**kw)
  if not ignore_status:
    resp.raise_for_status()
  return resp


def async2sync(func: Callable) -> Callable:
  """Превратить асинхронную функцию в синхронную | `asyncio`, `concurrent`"""
  import asyncio
  import concurrent.futures
  pool = concurrent.futures.ThreadPoolExecutor()

  def wrapper(*args, **kwargs):
    return pool.submit(asyncio.run, func(*args, **kwargs)).result()
  return wrapper


def get_my_ip() -> str:
  """Получить глобальный IP | `requests`"""
  with sync_request("GET", "https://api.ipify.org?format=json") as resp:
    ip = resp.json()["ip"]
  return ip


def is_async(func: Callable) -> bool:
  """Является ли функция асинхронной | `inspect`"""
  import inspect
  return inspect.iscoroutinefunction(func)


def is_sync(func: Callable) -> bool:
  """Является ли функция синхронной | `inspect`"""
  return not is_async(func)


def middleware(cls: type[MiddlewareBase]):
  """middleware для функции в виде декоратора"""
  def decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      return cls(func)._run(args, kwargs)
    return wrapper
  return decorator


def randfloat(min: float, max: float = None) -> float:
  """Случайное число | `random`"""
  from random import random
  if max == None:
    max = min
    min = 0
  return min + (random() * (max - min))


def randstr(length: int, symbols: str = "0123456789abcdefghijklmnopqrstuvwxyz") -> str:
  """Случайная строка | `random`"""
  from random import choice
  t = ""
  while len(t) < length:
    t += choice(symbols)
  return t


def riop(**p_kw):
  """Запустить функцию в отдельном процессе | `multiprocessing`"""
  import multiprocessing
  p_kw.setdefault("daemon", False)

  def decorator(func):
    p_kw["target"] = func

    @wraps(func)
    def wrapper(*args, **kwargs) -> multiprocessing.Process:
      p_kw["args"] = args
      p_kw["kwargs"] = kwargs
      p = multiprocessing.Process(**p_kw)
      p.start()
      return p
    return wrapper
  return decorator


def riot(**t_kw):
  """Запустить функцию в отдельном потоке | `threading`"""
  import threading
  t_kw.setdefault("daemon", False)

  def decorator(func):
    t_kw["target"] = func

    @wraps(func)
    def wrapper(*args, **kwargs) -> threading.Thread:
      t_kw["args"] = args
      t_kw["kwargs"] = kwargs
      t = threading.Thread(**t_kw)
      t.start()
      return t
    return wrapper
  return decorator


def sync_download_file(url: str, path: str, *, cb_end=return_None, cb_progress=return_None, cb_start=return_None, chunk_size: int = 1024, delete_on_error: bool = True, **kw) -> int:
  """Синхронная функция для скачивания файла | `requests`"""
  kw.setdefault("method", "GET")
  kw["stream"] = True
  kw["url"] = url
  with sync_request(**kw) as resp:
    if callable(getattr(path, "write", None)):
      f: IO[bytes] = path
    else:
      f = open(path, "wb")
    with f:
      size = 0
      cb_start(f, resp, size)
      try:
        for chunk in resp.iter_content(chunk_size):
          size += f.write(chunk)
          cb_progress(f, resp, size)
      except:
        f.close()
        if delete_on_error:
          if os.path.isfile(path):
            os.remove(path)
        raise
      cb_end(f, resp, size)
  return size


download_file = sync_download_file


def sync_request(method: str, url: str, *, ignore_status: bool = False, session=None, **kw):
  """Синхронный HTTP запрос | `requests`"""
  try:
    import requests
  except ImportError as err:
    try:
      from pip._vendor import requests
    except ImportError:
      raise err
  resp = None
  if isinstance(method, requests.Response):
    resp = method
  if isinstance(url, requests.Response):
    if not resp is None:
      raise TypeError("Only one argument can be `Response`")
    resp = url
  if resp is None:
    if session is None:
      session = requests.Session()
    kw["method"] = method
    kw["url"] = url
    resp = session.request(**kw)
  if not ignore_status:
    resp.raise_for_status()
  return resp


request = sync_request


def sync2async(func: Callable) -> Callable:
  """Превратить синхронную функцию в асинхронную"""
  async def wrapper(*args, **kwargs):
    return func(*args, **kwargs)
  return wrapper


def uuid(format=None, *args, **kwargs) -> str:
  """Сгенерировать UUID | `uuid`"""
  import uuid
  if format is None:
    cls = uuid.uuid4
  else:
    cls = getattr(uuid, f"uuid{format}")
  return str(cls(*args, **kwargs))


def timedelta(time: Union[int, float, dict]):
  """Превратить число/словарь в `timedelta` | `datetime`"""
  import datetime
  if isinstance(time, datetime.timedelta):
    return time
  if isinstance(time, dict):
    return datetime.timedelta(**time)
  return datetime.timedelta(seconds=time)


def shebang_code(code: str, *, exe_name: Union[None, str] = None, exe_path: Union[None, str] = None, none_if_no_changes: bool = False, use_env: bool = True) -> Union[None, str]:
  """Вставить/заменить шебанг в коде. Если указать имя интерпретатора, путь будет найден с помощью `shutil.which`"""
  if (exe_name is None) and (exe_path is None):
    raise TypeError("Specify exe_name or exe_path")
  if not exe_name is None:
    if not exe_path is None:
      raise TypeError("exe_name and exe_path cannot be used together")
  if exe_path is None:
    if use_env:
      exe_path = "/bin/env " + exe_name
    else:
      import shutil
      exe_path = shutil.which(exe_name)
      if exe_path is None:
        raise Exception("Command %s not found in PATH" % exe_name)
  lines = code.split("\n")
  if none_if_no_changes:
    if lines[0] == "#!" + exe_path:
      return None
  while len(lines[0].strip()) == 0 or lines[0].startswith("#!"):
    del lines[0]
  return "#!" + exe_path + "\n" + "\n".join(lines)


def shebang_file(path: str, **kw) -> int:
  """Вставить/заменить шебанг в файле кода"""
  kw["code"] = ms.file.read(path)
  kw["none_if_no_changes"] = True
  result = shebang_code(**kw)
  if result is None:
    return 0
  return ms.file.write(path, result)


def remove_ANSI(text: str) -> str:
  """Убрать ANSI коды из текста | `re`"""
  cache_id = "remove_ANSI", 0
  if not cache_id in cache:
    import re
    cache[cache_id] = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
  return cache[cache_id].sub("", text)


OnlyOneInstanceError = ms.types.OnlyOneInstanceError


class OnlyOneInstance:
  """Запретить запуск одной программы несколько раз | `tempfile`, `fcntl`"""
  _win = sys.platform == "win32"

  def __init__(self, name: str = "main", lock_path: str = None):
    import tempfile
    self.name: str = name
    self.running = False
    if lock_path is None:
      lock_path = tempfile.gettempdir() + "/" + ms.MAIN_FILE.replace(":", "").replace("/", "_") + "." + name + ".lock"
    self.lock = ms.path.Path(lock_path, use_cache=False)
    if self._win:
      flags = os.O_CREAT | os.O_EXCL | os.O_RDWR

      def _enter():
        try:
          if os.path.exists(self.lock.path):
            os.unlink(self.lock.path)
          self.fd = os.open(self.lock.path, flags)
        except OSError as err:
          if err.errno == 13:
            raise OnlyOneInstanceError()
          raise

      def _exit():
        os.close(self.fd)
        os.unlink(self.lock.path)
    else:
      import fcntl
      flags = fcntl.LOCK_EX | fcntl.LOCK_NB

      def _enter():
        self.fp = open(self.lock.path, "w")
        self.fp.flush()
        try:
          fcntl.lockf(self.fp, flags)
        except IOError:
          raise OnlyOneInstanceError()

      def _exit():
        fcntl.lockf(self.fp, fcntl.LOCK_UN)
        self.fp.close()
        if os.path.exists(self.lock.path):
          os.unlink(self.lock.path)
    self._enter = _enter
    self._exit = _exit

  def __enter__(self):
    if self.running:
      raise OnlyOneInstanceError()
    self._enter()
    self.running = True
    try:
      self.on_enter()
    except Exception:
      pass
    return self

  def __exit__(self, a, b, c):
    if not self.running:
      return
    self._exit()
    self.running = False
    self.lock.delete()
    self.on_exit()

  @classmethod
  def wrap_func(cls, **kw):
    def deco(func):
      ooi = cls(**kw)

      def wrapper(*args, **kwargs):
        with ooi:
          return func(*args, **kwargs)
      return wrapper
    return deco

  def on_enter(self):
    pass

  def on_exit(self):
    pass


def multi_and(*values: bool) -> bool:
  for i in values:
    if not i:
      return False
  return True


def multi_or(*values: bool) -> bool:
  for i in values:
    if i:
      return True
  return False


def is_int(value: float) -> bool:
  return value == int(value)


def get_self_module(__name__: str):
  return sys.modules[__name__]


def check_programs(*progs: str, raise_error: bool = True) -> list[str]:
  """Проверить наличие программ в `$PATH` | `shutil`"""
  from shutil import which
  failed = []
  for i in progs:
    if which(i) is None:
      if not i in failed:
        failed.append(i)
  failed.sort()
  if raise_error:
    if len(failed) > 0:
      if len(failed) == 1:
        raise OSError("Failed to find program " + failed[0] + " in $PATH")
      raise OSError("Failed to find programs " + (", ".join(failed)) + " in $PATH")
  return failed


def handle_exception(on_exception=return_None, reraise: bool = True, exc_type: type[BaseException] = Exception):
  """Обернуть функцию в `try`/`except`"""
  def deco(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      try:
        return func(*args, **kwargs)
      except exc_type as exc:
        on_exception(exc)
        if reraise:
          raise
    return wrapper
  return deco


def disable_warnings():
  """Отключить модуль `warnings`"""
  import warnings
  warnings.warn = return_None


def is_instance_of_one(obj, *classes: type) -> bool:
  """Это экземпляр одного из классов?"""
  for i in classes:
    if isinstance(obj, i):
      return True
  return False


def is_instance_of_all(obj, *classes: type) -> bool:
  """Это экземпляр всех классов?"""
  for i in classes:
    if not isinstance(obj, i):
      return False
  return True


def restore_deprecated(force: bool = False):
  """Восстановить устаревшие функции"""
  restored = []

  def setmethod(cls, name):
    def deco(func):
      if force or not hasattr(cls, name):
        try:
          setattr(cls, name, func)
          restored.append((cls, name, func))
        except Exception:
          pass
      return func
    return deco
  if "datetime" in sys.modules:
    from datetime import datetime

    @setmethod(datetime, "utcnow")
    def utcnow():
      return datetime.now(datetime.UTC)
  if "locale" in sys.modules:
    import locale

    @setmethod(locale, "resetlocale")
    def resetlocale(category=locale.LC_ALL):
      locale.setlocale(category, "")
  return restored


def main_func(_name_: str, exit: bool = True):
  def deco(func):
    if _name_ == "__main__":
      try:
        result = func()
      except KeyboardInterrupt:
        import signal
        result = signal.SIGINT.value
      if exit:
        if isinstance(result, int):
          sys.exit(result)
        sys.exit()
    return func
  return deco


class decorators:
  def __init__(self):
    raise NotImplementedError()

  @staticmethod
  def append(obj: list):
    """`obj.append(func)`"""
    def deco(func):
      obj.append(func)
      return func
    return deco

  @staticmethod
  def setattr(obj: type, name):
    """`setattr(obj,name,func)`"""
    def deco(func):
      setattr(obj, name, func)
      return func
    return deco

  @staticmethod
  def setitem(obj: dict | list, index):
    """`obj[index]=func`"""
    def deco(func):
      obj[index] = func
      return func
    return deco


def fassert(value: bool, text: str = None):
  """`assert` независимо от `__debug__`"""
  if value:
    return value
  if text is None:
    raise AssertionError()
  raise AssertionError(text)


def add2pythonpath(dir: str, to_begin: bool = False):
  """Добавить папку в `sys.path`, если её там нет"""
  dir = ms.path.path2str(dir, to_abs=True)
  if not dir in sys.path:
    if to_begin:
      sys.path.insert(0, dir)
    else:
      sys.path.append(dir)


def run_pip(*args: str, internal: bool = False, **kw):
  """Запустить `pip`. Встроенный запуск может завершить работу всей программы! | `pip`"""
  if getattr(sys, "frozen", False):
    raise RuntimeError("The environment is frozen")
  from pip._internal.exceptions import PipError
  if internal:
    from pip._internal.cli.main_parser import parse_command
    from pip._internal.commands import create_command
    cmd_name, cmd_args = parse_command(args)
    kw.setdefault("isolated", "--isolated" in cmd_args)
    command = create_command(cmd_name, **kw)
    if command.main(cmd_args):
      raise PipError()
  else:
    import subprocess
    kw["args"] = [sys.executable, "-m", "pip"] + list(args)
    if subprocess.call(**kw):
      raise PipError()


def check_modules(*modules: str, _c: list[str] = None, _m: list[str] = None) -> list[str]:
  """Проверить наличие модулей. Не проверяет возможность импорта. Возвращает список отсутствующих модулей"""
  import pkg_resources
  checked = [] if _c is None else _c
  missing = [] if _m is None else _m
  for i in modules:
    req = i if isinstance(i, pkg_resources.Requirement) else pkg_resources.Requirement.parse(i)
    if not req.name in checked:
      try:
        dist = pkg_resources.get_distribution(req)
        checked.append(req.name)
        check_modules(*dist.requires(), _c=checked, _m=missing)
      except pkg_resources.DistributionNotFound:
        req_str = str(req)
        if not req_str in missing:
          missing.append(req_str)
  return missing


def auto_install_modules(*modules, print: bool | str = False, **pip_kw):
  """Автоматически установить недостающие модули | `pip`, `pkg_resources`"""
  # Проверить возможность установки
  import pkg_resources
  from pip._internal.exceptions import PipError
  if getattr(sys, "frozen", False):
    raise RuntimeError("The environment is frozen")
  _ = PipError, pkg_resources  # Убрать предупреждение "модуль не используется"
  missing = check_modules(*modules)
  if len(missing) > 0:
    if print:
      text = print if isinstance(print, str) else "Выполняется установка модулей %(modules)s, подождите немного"
      builtins.print(text % {"modules": ", ".join(missing)}, file=sys.stderr)
    args = ["install", "-U"] + missing
    run_pip(*args, **pip_kw)


def http_check_range_support(url: str, **kw) -> bool:
  """Проверить возможность скачать файл по частям"""
  kw.setdefault("headers", {})
  kw["headers"]["Range"] = "bytes=0-1"
  kw["method"] = "HEAD"
  kw["stream"] = True
  kw["url"] = url
  with sync_request(**kw) as resp:
    if resp.status_code == 206:
      return True
    if "bytes" in resp.headers.get("Accept-Ranges", "").lower():
      return True
    return False


class MultiContext:
  def __init__(self, *objects):
    self.enter_handlers = []
    self.exit_handlers = []
    self.suppress_exc: None | type[BaseException] | set[type[BaseException]] = None
    self.add_objects(*objects)

  def __enter__(self):
    for i in self.enter_handlers:
      i()

  def __exit__(self, etype, exc, etb):
    for i in self.exit_handlers:
      try:
        i(etype, exc, etb)
      except Exception:
        pass
      from contextlib import suppress
    if not self.suppress_exc is None:
      if isinstance(self.suppress_exc, (list, set)):
        self.suppress_exc = tuple(self.suppress_exc)
      if isinstance(exc, self.suppress_exc):
        return True

  def _cl(self, l: list):
    if len(l) == 1:
      if isinstance(l[0], (list, set, tuple)):
        return list(l[0])
    return l

  def add_enter_handlers(self, *handlers):
    for i in self._cl(handlers):
      self.enter_handlers.append(i)

  def add_exit_handlers(self, *handlers):
    for i in self._cl(handlers):
      self.exit_handlers.append(i)

  def add_objects(self, *objects):
    for i in self._cl(objects):
      self.enter_handlers.append(getattr(i, "__enter__", return_None))
      self.exit_handlers.append(getattr(i, "__exit__", return_None))


def generator2list(list=builtins.list):
  """Превратить функцию с `yield` в функцию, которая возвращает список"""
  def deco(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      return list(func(*args, **kwargs))
    return wrapper
  return deco


def mini_log(msg: str, *values, **kw):
  """Напечатать текст в `stderr` с форматированием через `%`"""
  kw.setdefault("file", sys.stderr)
  if len(values) == 0:
    return print(msg, **kw)
  if len(values) == 1:
    values = values[0]
  print(msg % values, **kw)


def print_stderr(*values, **kw):
  """Тот же `print`, но всегда в `stderr`"""
  kw["file"] = sys.stderr
  print(*values, **kw)


def setattr_if_not_exists(obj, name: str, value, if_None=False):
  """Установить атрибут, если его нет"""
  if ((getattr(obj, name, None) is None) if if_None else (not hasattr(obj, name))):
    setattr(obj, name, value)


def call(func, args=[], kwargs={}):
  """Вызвать функцию. Если поставить как декоратор, функция будет вызвана сразу после определения и вместо неё будет сохранён результат её работы"""
  return func(*args, **kwargs)



"""Работа только с папками"""
import builtins
import os
from .core import ms
from .path import Path, PATH_TYPES, path2str
from typing import *
# 2.0.0


class NotADirError(Exception):
  pass


def _check(path, **kw) -> str:
  kw["path"] = path
  path = path2str(**kw)
  if os.path.exists(path):
    if not os.path.isdir(path):
      raise NotADirError(path)
  return path


def create(path: PATH_TYPES, force: bool = False, *, _exists: set[str] = None, **kw):
  """Создать папку если её не существует"""
  exists = set() if _exists is None else _exists
  path = path2str(path)
  if path in exists:
    return ms.path.Path(path)
  if os.path.isdir(path):
    exists.add(path)
    return ms.path.Path(path)
  if force:
    if os.path.isfile(path):
      ms.path.delete(path)
  kw["name"] = _check(path)
  os.makedirs(**kw)
  exists.add(path)
  return ms.path.Path(path)


def _list_filter(path: Path, *, exts: Iterable[str] = None, func: Callable[[Path], bool] = None, links: bool = None, type: str = None):
  if not exts is None:
    if not path.ext.lower() in exts:
      return False
  if not func is None:
    if not func(path):
      return False
  if not links is None:
    if path.is_link != links:
      return False
  if not type is None:
    if path.type != type:
      return False
  return True


def list_iter(path: PATH_TYPES = ".", *, exts: Iterable[str] = None, **kw):
  """Список содержимого папки (итератор)"""
  path = _check(path)
  kw["exts"] = None if exts is None else [(i if i.startswith(".") else "." + i).lower() for i in exts]
  for i in os.listdir(path):
    i = Path(path + "/" + i)
    if _list_filter(i, **kw):
      yield i


def list(path: PATH_TYPES = ".", **kw) -> builtins.list[Path]:
  """Список содержимого папки"""
  return builtins.list(list_iter(path, **kw))


def copy(path: PATH_TYPES, dest: PATH_TYPES, **kw):
  """Копировать папку"""
  kw["dest"] = dest
  kw["path"] = _check(path)
  return ms.path.copy(**kw)


def delete(path: PATH_TYPES, **kw):
  """Удалить папку с содержимым"""
  kw["path"] = _check(path)
  return ms.path.delete(**kw)


def in_dir(path: str, dir: str, **kw) -> bool:
  """Находится ли папка в указанной папке"""
  kw["dir"] = dir
  kw["path"] = _check(path)
  return ms.path.in_dir(**kw)


def link(path: PATH_TYPES, dest: PATH_TYPES, **kw):
  """Сделать символическую ссылку на папку"""
  kw["dest"] = dest
  kw["path"] = _check(path)
  return ms.path.link(**kw)


def move(path: PATH_TYPES, dest: PATH_TYPES, **kw):
  """Переместить папку"""
  kw["dest"] = dest
  kw["path"] = _check(path)
  return ms.path.move(**kw)


def rename(path: PATH_TYPES, name: PATH_TYPES, **kw):
  """Переименовать папку"""
  kw["name"] = name
  kw["path"] = _check(path)
  return ms.path.rename(**kw)


class TempDir:
  """Временная папка, которая будет удалена по окончании использования (используйте через `with`)"""

  def __init__(self, path: PATH_TYPES, *, create: bool = True, ignore_error: bool = True):
    self.ignore_error = ignore_error
    self.path: str = path2str(path, True)
    if create:
      self.create()

  def __enter__(self):
    self.create()
    return self

  def __exit__(self, type, value, trace):
    if not value is None:
      if not self.ignore_error:
        return
    self.delete()

  def create(self):
    """Создать папку если она не существует"""
    ms.dir.create(self.path)

  def delete(self):
    """Удалить папку вместе с содержимым"""
    ms.path.delete(self.path)


def recursive_list_iter(path: PATH_TYPES, follow_links: bool = False, *, on_error=None, top_down=True, exts: Iterable[str] = None, **kw):
  """Рекурсивный список содержимого папки (итератор)"""
  kw["exts"] = None if exts is None else [(i if i.startswith(".") else "." + i).lower() for i in exts]
  for dirpath, dirnames, filenames in os.walk(path, followlinks=follow_links, onerror=on_error, topdown=top_down):
    for i in dirnames:
      i = Path(dirpath + "/" + i)
      if _list_filter(i, **kw):
        yield i
    for i in filenames:
      i = Path(dirpath + "/" + i)
      if _list_filter(i, **kw):
        yield i


def recursive_list(path: PATH_TYPES = ".", follow_links: bool = False, **kw) -> builtins.list[Path]:
  """Рекурсивный список содержимого папки"""
  return builtins.list(recursive_list_iter(path, follow_links, **kw))


cp = copy
ln = link
mv = move
rm = delete
rn = rename



import hashlib
import os
import shutil
import subprocess
import sys
from MainShortcuts2 import ms
from zipfile import ZipFile, ZipInfo
interpreters: dict[str, list[str]] = {}


def reg_interpreter(command: str | list[str], *extensions: str):
  if isinstance(command, str):
    command = [command]
  for ext in extensions:
    if not ext.startswith("."):
      ext = "." + ext
    interpreters[ext] = command


def get_command(file: str) -> list[str]:
  _, ext = os.path.splitext(file)
  return interpreters.get(ext, []) + [file]


def _get_cache_dir():
  if sys.platform == "win32":
    import winreg
    dir, type = winreg.QueryValueEx(winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders"), "Local AppData")
    return ms.path.Path(dir + "/MainPlay TG/MainShortcuts2/ms2app_cache")
  if sys.platform == 'darwin':
    return ms.path.Path(os.path.expanduser("~/Library/Caches") + "/ms2app_cache")
  return ms.path.Path(os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache")) + "/ms2app_cache")


CACHE_DIR = _get_cache_dir()


class LaunchError(Exception):
  pass


class Metadata(ms.ObjectBase):
  FILENAME = "ms2app.json"
  FORMAT = 1
  KEYS = (
      "author",
      "custom",
      "description",
      "download_url",
      "executable",
      "license",
      "name",
      "update_url",
      "version_id",
      "version",
  )

  def __init__(self, author: str, name: str, *,
               custom: dict = None,
               description: str = None,
               download_url: str = None,
               executable: str = "__main__.py",
               license: str = None,
               update_url: str = None,
               version_id: int = -1,
               version: str = None,
               ):
    self.author: str = author
    """Автор приложения"""
    self.download_url: None | str = download_url
    """Ссылка для скачивания `.ms2app`. Обязательна при проверке обновлений"""
    self.custom: dict = {} if custom is None else custom
    """Пользовательские метаданные"""
    self.description: None | str = description
    """Описание приложения"""
    self.executable: str = executable
    """Название главного файла приложения"""
    self.license: None | str = license
    """Лицензия"""
    self.name: str = name
    """Название приложения"""
    self.update_url: None | str = update_url
    """Ссылка на JSON файл с информацией для скачивания/обновления"""
    self.version_id: int = version_id
    """ID версии, используется для проверки обновлений"""
    self.version: str = str(version_id) if version is None else version
    """Название версии"""
  @classmethod
  def from_dict(cls, data):
    return cls(**data)

  @classmethod
  def from_json(cls, text, **json_kw):
    return cls.from_dict(ms.json.decode(text, **json_kw))

  def to_dict(self) -> dict:
    data = {"format": self.FORMAT}
    for k in self.KEYS:
      data[k] = getattr(self, k, None)
    return data

  def to_json(self, **kw) -> str:
    return ms.json.encode(self.to_dict(), **kw)

  def to_jsonb(self, **kw) -> bytes:
    return self.to_json(**kw).encode(ms.encoding)

  def check_updates(self) -> None | str:
    """Проверить наличие обновлений. Возвращает `None` или ссылку на скачивание `.ms2app`"""
    if not self.update_url:
      return
    try:
      with ms.utils.request("GET", self.update_url) as resp:
        data: dict = resp.json()
        if self.name != data["name"]:
          print("Warning: The name of the new version of the program is different (%s -> %s)" % (self.name, data["name"]), file=sys.stderr)
        if self.version_id > data["version_id"]:
          return data.get("download_url")
    except Exception as exc:
      print("Check for updates error: %r" % exc, file=sys.stderr)


class App(ms.ObjectBase):
  def __init__(self, file: str, password: bytes = None):
    self._platform = None
    self._sha256 = None
    self.file = ms.path.Path(file)
    self.password = password.encode("utf-8") if isinstance(password, str) else password
    with self.open_zip("r") as zip:
      self.meta = Metadata.from_json(zip.read(Metadata.FILENAME))

  def _get_platform(self) -> str:
    if self._platform is None:
      with self.open_zip() as zip:
        namelist = zip.namelist()
      if "app-%s/%s" % (sys.platform, self.meta.executable) in namelist:
        self._platform = sys.platform
      elif "app-any/%s" % self.meta.executable in namelist:
        self._platform = "any"
      else:
        raise LaunchError("This app is not supported on your system (%s)" % sys.platform)
    return self._platform

  @property
  def sha256(self) -> str:
    if self._sha256 is None:
      with open(self.file, "rb") as f:
        hash = hashlib.sha256()
        for i in f:
          hash.update(i)
      self._sha256 = hash.hexdigest()
    return self._sha256

  def open_zip(self, mode: str = "r", **kw) -> ZipFile:
    kw["mode"] = mode
    zip = ZipFile(self.file.path, **kw)
    if self.password:
      zip.setpassword(self.password)
    else:
      if zip.comment.lower().startswith(b"password: "):
        zip.setpassword(zip.comment[10:zip.comment.index(b"\n")])
    return zip

  def _extract_file(self, zip: ZipFile, src: ZipInfo, dest, created_dirs):
    if not isinstance(src, ZipInfo):
      src = zip.getinfo(src)
    if not isinstance(dest, ms.path.Path):
      dest = ms.path.Path(dest)
    if src.is_dir():
      return
    if dest.exists:
      return
    if not dest.parent_dir in created_dirs:
      if not os.path.exists(dest.parent_dir):
        os.makedirs(dest.parent_dir)
      created_dirs.add(dest.parent_dir)
    with zip.open(src, "r") as source:
      with open(dest, "wb") as target:
        shutil.copyfileobj(source, target)

  def _check_prefix(self, filename: str):
    for prefix in ("app-%s/" % self._get_platform(), "assets/", "icon.", "libs/"):
      if filename.startswith(prefix):
        return True
    return False

  def unpack(self, extract_dir: str = None):
    """Распаковать архив в папку кеша"""
    created_dirs = set()
    if extract_dir is None:
      extract_dir = CACHE_DIR + "/app-" + self.sha256 + "/"
    elif not extract_dir.endswith("/"):
      extract_dir += "/"
    with self.open_zip() as zip:
      for file in zip.filelist:
        if self._check_prefix(file.filename):
          dest = ms.path.Path(extract_dir + file.filename)
          self._extract_file(zip, file, dest, created_dirs)
      self._extract_file(zip, Metadata.FILENAME, extract_dir + Metadata.FILENAME, created_dirs)

  def start(self, args: list[str] = None, **kw):
    """запустить приложение"""
    from . import _ms2app_regs
    app_dir = CACHE_DIR + "/app-" + self.sha256 + "/app-" + self._get_platform()
    app_exe = app_dir + "/" + self.meta.executable
    kw.setdefault("env", os.environ.copy())
    kw["args"] = get_command(app_exe)
    kw["env"]["MS2APP_APP_DIR"] = app_dir
    kw["env"]["MS2APP_APP_EXE"] = app_exe
    kw["env"]["MS2APP_ZIP_FILE"] = self.file.path
    if args:
      kw["args"] += args
    return subprocess.Popen(**kw)

  def start_call(self, args: list[str] = None, timeout: float = None, **kw):
    """Запустить приложение и ожидать завершения"""
    with self.start(args, **kw) as p:
      try:
        return p.wait(timeout=timeout)
      except:
        p.kill()
        raise


@ms.utils.main_func(__name__)
def main(args=None, **kw):
  if args is None:
    from argparse import ArgumentParser
    argp = ArgumentParser("ms2-app", description="Управление приложениями ms2app")
    subp = argp.add_subparsers(dest="command", required=True)
    # download
    subp_download = subp.add_parser("download", help="скачать приложение")
    subp_download.add_argument("-O", "--output", help="путь сохранения")
    subp_download.add_argument("url", help="ссылка для скачивания")
    # run
    subp_run = subp.add_parser("run", help="запустить приложение")
    subp_run.add_argument("--password", help="пароль от файла")
    subp_run.add_argument("-b", "--no-wait", action="store_true", help="запустить приложение в фоне")
    subp_run.add_argument("args", nargs="+", help="путь к файлу ms2app или zip и аргументы запуска")
    # show
    subp_show = subp.add_parser("show", help="показать информацию о приложении")
    subp_show.add_argument("--no-check-updates", action="store_true", help="не проверять наличие обновлений")
    subp_show.add_argument("--password", help="пароль от файла")
    subp_show.add_argument("file", help="путь к файлу ms2app или zip")
    # upgrade
    subp_upgrade = subp.add_parser("upgrade", help="обновить приложение")
    subp_upgrade.add_argument("--password", help="пароль от файла")
    subp_upgrade.add_argument("-O", "--output", help="путь сохранения")
    subp_upgrade.add_argument("file", help="путь к файлу ms2app или zip")
    args = argp.parse_args()
  if args.command == "download":
    kw.setdefault("url", args.url)
    if args.output:
      kw.setdefault("path", args.output)
    else:
      from urllib.parse import ParseResult, urlparse
      parsed_url: ParseResult = urlparse(args.url)
      kw.setdefault("path", parsed_url.path.split("/")[-1])
    ms.utils.download_file(**kw)
    return
  if args.command == "run":
    app = App(args.args.pop(0), args.password)
    if args.no_wait:
      return app.start(args.args)
    return app.start_call(args.args)
  if args.command == "show":
    app = App(args.file, args.password)
    data = app.meta.to_dict()
    data["file"] = app.file.path
    for k in list(data):
      v = data[k]
      if isinstance(v, bool):
        data[k] = "да" if v else "нет"
      elif not v:
        data[k] = "(отсутствует)"
    print("Приложение %(name)s от %(author)s" % data)
    print("Версия: %(version)s (%(version_id)s)" % data)
    print("Описание: %(description)s" % data)
    print("Файл: %(file)s" % data)
    print("Лицензия: %(license)s" % data)
    if not args.no_check_updates:
      data["update"] = app.meta.check_updates()
      if data["update"]:
        print("Доступно обновление! Можете скачать его командой\nms2-app upgrade %(file)s\nИли по ссылке %(update)s" % data)
    return
  if args.command == "upgrade":
    app = App(args.file, args.password)
    kw.setdefault("url", app.meta.check_updates())
    if not kw["url"]:
      return print("Нет доступных обновлений")
    if args.output:
      kw.setdefault("path", args.output)
    else:
      kw.setdefault("path", app.file)
    print("Скачивание обновления...")
    ms.utils.download_file(**kw)
    print("Обновление завершено")
    return



import argparse
import hashlib
import os
import shlex
import sys
from MainShortcuts2 import ms
HASH_SUFFIX = ".MS2_hash"
HASH_TYPES = ["blake2b", "blake2s", "md5", "sha1", "sha224", "sha256", "sha384", "sha3_224", "sha3_256", "sha3_384", "sha3_512", "sha512"]


class _NoPbar(ms.ObjectBase):
  finish = ms.utils.return_None
  start = ms.utils.return_None
  update = ms.utils.return_None
  value = 0


def get_pbar(filename: str, enabled: bool, **kw):
  if not enabled:
    return _NoPbar()
  try:
    import progressbar
  except Exception:
    return _NoPbar()
  kw.setdefault("max_error", False)
  kw.setdefault(
      "widgets",
      [
          filename,
          " ",
          progressbar.Percentage(),
          progressbar.GranularBar(
              left="(",
              right=")",
          ),
          progressbar.FileTransferSpeed(),
          " ",
          progressbar.ETA(
              format_finished="%(elapsed)8s",
              format_na="     N/A",
              format_not_started="--:--:--",
              format_zero="00:00:00",
              format="%(eta)8s",
          )
      ]
  )
  return progressbar.ProgressBar(**kw)


class Format1:
  def __init__(self, *, file_size: int, hash_hex: str, hash_type: str):
    assert hash_type in HASH_TYPES
    self.file_size = file_size
    self.hash_hex = hash_hex
    self.hash_type = hash_type

  @classmethod
  def from_dict(cls, data: dict, **kw):
    kw.setdefault("file_size", data["file"]["size"])
    kw.setdefault("hash_hex", data["hash"]["hex"])
    kw.setdefault("hash_type", data["hash"]["type"])
    return cls(**kw)

  @classmethod
  def generate(cls, path: str, enable_pbar: bool, **kw):
    file = ms.path.Path(path)
    kw.setdefault("hash_type", "sha512")
    hash: hashlib._Hash = getattr(hashlib, kw["hash_type"])()
    kw["file_size"] = file.size
    with open(path, "rb") as f:
      with get_pbar(file.full_name, enable_pbar, max_value=file.size) as pbar:
        pbar.start()
        for i in f:
          hash.update(i)
          pbar.update(pbar.value + len(i))
    kw["hash_hex"] = hash.hexdigest()
    return cls(**kw)

  def to_dict(self):
    data = {}
    data["file"] = {"size": self.file_size}
    data["format"] = "MainShortcuts2_hash_v1"
    data["hash"] = {"hex": self.hash_hex, "type": self.hash_type}
    return data


def hash_gen(args: argparse.Namespace = None):
  if args is None:
    argp = argparse.ArgumentParser("ms2-hash_gen", description="создание контрольной суммы для файла")
    argp.add_argument("files", nargs="+", help="пути к файлам")
    argp.add_argument("-b", "--bar", action="store_true", help="показывать прогрессбар (нужен модуль progressbar2)")
    argp.add_argument("-f", "--force", action="store_true", help="перезаписывать существующие хеши")
    argp.add_argument("-t", "--type", choices=HASH_TYPES, default="sha512", help="тип контрольной суммы")
    args = argp.parse_args()
  completed = []
  for file in args.files:
    while file.lower().endswith(HASH_SUFFIX.lower()):
      file = file[:0 - len(HASH_SUFFIX)]
    if file in completed:
      continue
    if os.path.isdir(file):
      print("Пропуск файла " + shlex.quote(file) + ": это папка", file=sys.stderr)
      completed.append(file)
      continue
    if os.path.isfile(file + HASH_SUFFIX):
      if not args.force:
        print("Пропуск файла " + shlex.quote(file) + ": хеш существует", file=sys.stderr)
        continue
    hash = Format1.generate(file, args.bar, hash_type=args.type)
    ms.json.write(file + HASH_SUFFIX, hash.to_dict())
    completed.append(file)


def hash_check(args: argparse.Namespace = None):
  if args is None:
    argp = argparse.ArgumentParser("ms2-hash_check", description="проверка размера и контрольной суммы файла")
    argp.add_argument("files", nargs="+", help="пути к файлам")
    argp.add_argument("-b", "--bar", action="store_true", help="показывать прогрессбар (нужен модуль progressbar2)")
    args = argp.parse_args()
  completed = []
  for file in args.files:
    while file.lower().endswith(HASH_SUFFIX.lower()):
      file = file[:0 - len(HASH_SUFFIX)]
    if file in completed:
      continue
    if os.path.isdir(file):
      print("Пропуск файла " + shlex.quote(file) + ": это папка", file=sys.stderr)
      completed.append(file)
      continue
    if not os.path.exists(file + HASH_SUFFIX):
      print("Ошибка: не найден файл " + shlex.quote(file + HASH_SUFFIX), file=sys.stderr)
      continue
    saved_hash = Format1.from_dict(ms.json.read(file + HASH_SUFFIX))
    if saved_hash.file_size != os.path.getsize(file):
      print("Ошибка: размер файла " + shlex.quote(file) + " не совпадает", file=sys.stderr)
      continue
    real_hash = Format1.generate(file, args.bar, hash_type=saved_hash.hash_type)
    if saved_hash.hash_hex == real_hash.hash_hex:
      print("Успех: файл " + shlex.quote(file) + " не изменён")
    else:
      print("Ошибка: файл " + shlex.quote(file) + " изменён")
    completed.append(file)
